from typer import Exit, Typer

from ..exc import HostedPiException
from . import arguments, options
from .sshkeys import keys_app
from .utils import get_pi, get_pis, print_error


ssh_app = Typer()
ssh_app.add_typer(keys_app, name="keys", no_args_is_help=True, help="SSH key management commands")


@ssh_app.command("command")
def do_command(
    name: arguments.server_name,
    ipv6: options.ipv6 = False,
    numeric: options.numeric = False,
    user: options.user = "root",
):
    """
    Get the SSH command to connect to a Raspberry Pi server
    """
    if numeric and not ipv6:
        print_error("--numeric is only supported with --ipv6")
        raise Exit(1)

    pi = get_pi(name)
    if pi is None:
        print_error(f"Pi '{name}' not found")
        raise Exit(1)
    try:
        if ipv6:
            print(pi.get_ipv6_ssh_command(numeric=numeric, user=user))
        else:
            print(pi.get_ipv4_ssh_command(user=user))
    except HostedPiException as exc:
        print_error(f"hostedpi error: {exc}")
        raise Exit(1)


@ssh_app.command("config")
def do_config(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
    ipv6: options.ipv6 = False,
    numeric: options.numeric = False,
    user: options.user = "root",
):
    """
    Get the SSH config to connect to one or more Raspberry Pi servers
    """
    if numeric and not ipv6:
        print_error("--numeric is only supported with --ipv6")
        raise Exit(1)

    pis = get_pis(names, filter)
    for pi in pis:
        try:
            if ipv6:
                print(pi.get_ipv6_ssh_config(numeric=numeric, user=user))
            else:
                print(pi.get_ipv4_ssh_config(user=user))
        except HostedPiException as exc:
            print_error(f"hostedpi error: {exc}")
            raise Exit(1)



from pathlib import Path
from typing import Annotated, Union

from typer import Option


server_name = Annotated[Union[str, None], Option(help="Name of the new Raspberry Pi server")]
model = Annotated[int, Option(help="Raspberry Pi Model", min=3, max=4)]
disk_size = Annotated[int, Option(help="Disk size in GB", min=10)]
memory = Annotated[Union[int, None], Option(help="Memory in GB", min=1, max=8)]
cpu_speed = Annotated[Union[int, None], Option(help="CPU speed in MHz", min=1500)]
os_image = Annotated[Union[str, None], Option(help="Operating system image")]
wait = Annotated[bool, Option(help="Wait and poll for status to be available before returning")]
ssh_key_path = Annotated[
    Union[Path, None], Option(help="Path to the SSH key to install on the Raspberry Pi servers")
]
ssh_import_github = Annotated[
    Union[list[str], None],
    Option("--github", "--gh", help="GitHub usernames to source SSH keys from"),
]
ssh_import_launchpad = Annotated[
    Union[list[str], None],
    Option("--launchpad", "--lp", help="Launchpad usernames to source SSH keys from"),
]
ipv6 = Annotated[bool, Option(help="Use the IPv6 connection method")]
numeric = Annotated[bool, Option("--numeric", "-n", help="Use numeric IPv6 address in SSH command")]
user = Annotated[str, Option("--username", "-u", help="Username for SSH connection")]
yes = Annotated[bool, Option("--yes", "-y", help="Proceed without confirmation")]
number = Annotated[Union[int, None], Option(help="Number of Raspberry Pi servers to create", min=1)]
full_table = Annotated[bool, Option(help="Show full table of Raspberry Pi server info")]
filter_pattern_pi = Annotated[
    Union[str, None], Option(help="Search pattern for filtering server names")
]
filter_pattern_images = Annotated[
    Union[str, None], Option(help="Search pattern for filtering image names")
]



import urllib.parse
from datetime import datetime, timezone
from functools import cached_property
from ipaddress import IPv6Address, IPv6Network
from time import sleep
from typing import Union

from pydantic import ValidationError
from requests import ConnectionError, HTTPError, Session
from structlog import get_logger

from .auth import MythicAuth
from .exc import (
    HostedPiNotAuthorizedError,
    HostedPiProvisioningError,
    HostedPiServerError,
    HostedPiUserError,
)
from .logger import log_request
from .models.mythic.responses import (
    PiInfo,
    PiInfoBasic,
    ProvisioningServer,
    SSHKeysResponse,
)
from .models.sshkeys import SSHKeySources
from .utils import (
    dedupe_ssh_keys,
    get_error_message,
    remove_imported_ssh_keys,
    remove_ssh_keys_by_label,
)


logger = get_logger()


class Pi:
    """
    The ``Pi`` class represents a single Raspberry Pi service in the Mythic Beasts Pi cloud.
    Initialising a ``Pi`` object does not provision a new Pi, rather initialisation is for internal
    construction only.

    There are two ways to get access to a ``Pi`` object: retrieval from the
    :attr:`~hostedpi.picloud.PiCloud.pis` dictionary; and the return value of
    :meth:`~hostedpi.picloud.PiCloud.create_pi` method.

    With a ``Pi`` object, you can access data about that particular Pi service, add SSH keys, reboot
    it, cancel it and more.

    .. note::
        The ``Pi`` class should not be initialised by the user, only internally within the module
    """

    def __init__(
        self,
        name: Union[str, None],
        *,
        info: PiInfoBasic,
        auth: Union[MythicAuth, None] = None,
        status_url: Union[str, None] = None,
    ):
        self._name = name
        self._model = info.model
        self._memory = info.memory
        self._cpu_speed = info.cpu_speed
        if auth is None:
            auth = MythicAuth()
        self._auth = auth
        self._api_url = str(auth.settings.api_url)
        self._cancelled = False
        self._info: Union[PiInfoBasic, PiInfo, None] = None
        self._last_fetched_info: Union[datetime, None] = None
        self._status_url: Union[str, None] = status_url

    def __repr__(self):
        if self._cancelled:
            return f"<Pi name={self.name} cancelled>"
        if self._info is None:
            return f"<Pi name={self.name} model={self.model}>"
        model = self.model_full if self.model_full else self.model
        return f"<Pi name={self.name} model={model}>"

    @property
    def session(self) -> Session:
        """
        The authenticated requests session used to communicate with the API
        """
        return self._auth.session

    @property
    def info(self) -> PiInfo:
        """
        The full Pi information as a :class:`~hostedpi.models.mythic.responses.PiInfo` object.
        Always fetches the latest information from the API when this is called (with a cache timeout
        of 10 seconds).

        :raises HostedPiUserError:
            If the Pi has not been initialised with a name yet, or if the name is ``None``

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server info

        :raises HostedPiProvisioningError:
            If there is an error retrieving the Pi information from the API because the Pi is still
            provisioning

        :raises HostedPiServerError:
            If there is another error retrieving the Pi information from the API
        """
        if self._info is None:
            self._get_info()
        return self._info

    @property
    def name(self) -> str:
        """
        The name of the Pi
        """
        return self._name

    @property
    def model(self) -> int:
        """
        The Pi's model (3 or 4)
        """
        return self._model

    @cached_property
    def model_full(self) -> Union[str, None]:
        """
        The Pi's model name (3B, 3B+ or 4B)
        """
        return self.info.model_full

    @property
    def memory_mb(self) -> Union[int, None]:
        """
        The Pi's RAM size in MB
        """
        return self._memory

    @property
    def memory_gb(self) -> Union[int, None]:
        """
        The Pi's RAM size in GB
        """
        return self._memory // 1024 if self._memory else None

    @property
    def cpu_speed(self) -> Union[int, None]:
        """
        The Pi's CPU speed in MHz
        """
        return self._cpu_speed

    @cached_property
    def disk_size(self) -> Union[int, None]:
        """
        The Pi's disk size in GB
        """
        return self.info.disk_size

    @cached_property
    def nic_speed(self) -> Union[int, None]:
        """
        The Pi's NIC speed in MHz
        """
        return self.info.nic_speed

    @property
    def status(self) -> str:
        """
        A string representing the Pi's current status (provisioning, booting, live, powered on or
        powered off).
        """
        self._get_info()
        if self.provision_status != "live":
            return f"Provisioning: {self.provision_status}"
        if self.info.is_booting:
            return f"Booting: {self.boot_progress}"
        if self.power:
            return "Powered on"
        return "Powered off"

    @property
    def boot_progress(self) -> str:
        """
        A string representing the Pi's boot progress. Can be ``booted``, ``powered off`` or a
        particular stage of the boot process if currently booting.
        """
        if self.info.boot_progress:
            return self.info.boot_progress
        return "booted" if self.power else "powered off"

    @property
    def initialised_keys(self) -> bool:
        """
        A boolean representing whether or not the Pi has been initialised with SSH keys
        """
        return self.info.initialised_keys

    @cached_property
    def ipv4_ssh_port(self) -> int:
        """
        The SSH port to use when connecting via the IPv4 proxy
        """
        return self.info.ssh_port

    @cached_property
    def ipv6_address(self) -> IPv6Address:
        """
        The Pi's IPv6 address as an :class:`~ipaddress.IPv6Address` object
        """
        return self.info.ipv6_address

    @cached_property
    def ipv6_network(self) -> IPv6Network:
        """
        The Pi's IPv6 network as an :class:`~ipaddress.IPv6Network` object
        """
        return self.info.ipv6_network

    @property
    def is_booting(self) -> bool:
        """
        A boolean representing whether or not the Pi is currently booting
        """
        return self.info.is_booting

    @cached_property
    def location(self) -> str:
        """
        The Pi's physical location (data centre)
        """
        return self.info.location

    @property
    def power(self) -> bool:
        """
        A boolean representing whether or not the Pi is currently powered on
        """
        return self.info.power

    @property
    def provision_status(self) -> str:
        """
        A string representing the provision status of the Pi. Can be "provisioning", "initialising"
        or "live".
        """
        return self.info.provision_status

    @property
    def hostname(self) -> str:
        """
        The hostname of the Pi
        """
        return f"{self.name}.hostedpi.com"

    @property
    def ipv4_ssh_hostname(self) -> str:
        """
        The hostname to use when connecting to the Pi over SSH using IPv4
        """
        return f"ssh.{self.hostname}"

    @property
    def ipv6_ssh_hostname(self) -> str:
        """
        The hostname to use when connecting to the Pi using SSH over IPv6
        """
        return self.hostname

    @property
    def ipv4_ssh_command(self) -> str:
        """
        The default SSH command required to connect to the Pi using SSH over IPv4
        """
        return self.get_ipv4_ssh_command()

    @property
    def ipv6_ssh_command(self) -> str:
        """
        The default SSH command required to connect to the Pi using SSH over IPv6
        """
        return self.get_ipv6_ssh_command()

    @property
    def ipv4_ssh_config(self) -> str:
        """
        A string containing the default IPv4 SSH config for the Pi. The contents could be added to
        an SSH config file for easy access to the Pi.
        """
        return self.get_ipv4_ssh_config()

    @property
    def ipv6_ssh_config(self) -> str:
        """
        A string containing the default IPv6 SSH config for the Pi. The contents could be added to
        an SSH config file for easy access to the Pi.
        """
        return self.get_ipv6_ssh_config()

    @property
    def url(self) -> str:
        """
        The http version of the hostedpi.com URL of the Pi.

        .. note::
            Note that a web server must be installed on the Pi for the URL to be resolvable.
        """
        return f"http://www.{self.hostname}"

    @property
    def url_ssl(self) -> str:
        """
        The https version of the hostedpi.com URL of the Pi.

        .. note::
            Note that a web server must be installed on the Pi for the URL to be resolvable, and an
            SSL certificate must be created.

            See https://letsencrypt.org/
        """
        return f"https://www.{self.hostname}"

    @property
    def ssh_keys(self) -> set[str]:
        """
        Retrieve the SSH keys on the Pi, or use assignment to update them. Property value is a set
        of strings. Assigned value should also be a set of strings, or None to unset.

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the Pi is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API

        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-piserversidentifierssh-key
        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}/ssh-key")
        response = self.session.get(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiProvisioningError(error) from exc
            raise HostedPiServerError(error) from exc

        data = SSHKeysResponse.model_validate(response.json())

        return dedupe_ssh_keys(data.keys)

    @ssh_keys.setter
    def ssh_keys(self, ssh_keys: Union[set[str], None]):
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-put-piserversidentifierssh-key
        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}/ssh-key")

        if ssh_keys is None:
            data = {"ssh_key": ""}
        else:
            data = {"ssh_key": "\r\n".join(dedupe_ssh_keys(ssh_keys))}

        response = self.session.put(url, json=data)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiProvisioningError(error) from exc
            raise HostedPiServerError(error) from exc

    def on(self, *, wait: bool = False) -> Union[bool, None]:
        """
        Power the Pi on. If *wait* is ``False`` (the default), return immediately. If *wait* is
        ``True``, wait until the power on request is completed, and return ``True`` on success, and
        ``False`` on failure.

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the server is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        self._power_on_off(on=True)
        if wait:
            while self.info.is_booting:
                sleep(10)
            return self.power

    def off(self):
        """
        Power the Pi off and return immediately

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the server is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        self._power_on_off(on=False)

    def reboot(self, *, wait: bool = False):
        """
        Reboot the Pi. If *wait* is ``False`` (the default), return ``None`` immediately. If *wait*
        is ``True``, wait until the reboot request is completed, and return ``True`` on success, and
        ``False`` on failure.

        .. note::
            Note that if *wait* is ``False``, you can poll for the boot status while rebooting by
            inspecting the properties :attr:`~hostedpi.pi.Pi.is_booting` and
            :attr:`~hostedpi.pi.Pi.boot_progress`.

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-post-piserversidentifierreboot
        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}/reboot")
        response = self.session.post(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            if response.status_code == 409:
                # The server is already being rebooted
                pass
            else:
                error = get_error_message(exc)
                if response.status_code == 403:
                    raise HostedPiNotAuthorizedError(error) from exc
                raise HostedPiServerError(error) from exc

        if wait:
            while self.info.is_booting:
                sleep(5)
            return self.power

    def cancel(self):
        """
        Unprovision the Pi server immediately

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the server is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-delete-piserversidentifier
        if self._cancelled:
            logger.warn("This Pi server is already cancelled", name=self.name)
            return

        # check if the server is still provisioning
        status = self.get_provision_status()
        if type(status) is not PiInfo:
            logger.warn("Cannot cancel a server that is still provisioning", name=self.name)
            return

        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}")
        response = self.session.delete(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiProvisioningError(error) from exc
            raise HostedPiServerError(error) from exc

        self._cancelled = True

    def get_ipv4_ssh_command(self, user: str = "root") -> str:
        """
        Construct an SSH command required to connect to the Pi using SSH over IPv4
        """
        return f"ssh -p {self.ipv4_ssh_port} {user}@{self.ipv4_ssh_hostname}"

    def get_ipv6_ssh_command(self, *, user: str = "root", numeric: bool = False) -> str:
        """
        Construct an SSH command required to connect to the Pi using SSH over IPv6
        """
        if numeric:
            return f"ssh {user}@[{self.ipv6_address.compressed}]"
        else:
            return f"ssh {user}@{self.ipv6_ssh_hostname}"

    def get_ipv4_ssh_config(self, user: str = "root") -> str:
        """
        Construct a string containing the IPv4 SSH config for the Pi. The contents could be added to
        an SSH config file for easy access to the Pi.
        """
        return f"""Host {self.name}
    user {user}
    port {self.ipv4_ssh_port}
    hostname {self.ipv4_ssh_hostname}
        """.strip()

    def get_ipv6_ssh_config(
        self,
        *,
        user: str = "root",
        numeric: bool = False,
    ) -> str:
        """
        Construct a string containing the SSH config for the Pi. The contents could be added to an
        SSH config file for easy access to the Pi.
        """
        hostname = self.ipv6_address.compressed if numeric else self.ipv6_ssh_hostname
        return f"""Host {self.name}
    user {user}
    hostname {hostname}
        """.strip()

    def add_ssh_keys(self, ssh_keys: SSHKeySources) -> set[str]:
        """
        Add SSH keys to the Pi from the specified sources.

        :type ssh_keys: SSHKeySources
        :param ssh_keys: The sources to find keys to add to the Pi

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the Pi is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        keys = ssh_keys.collect()
        if keys:
            self.ssh_keys |= keys
        return self.ssh_keys

    def unimport_ssh_keys(
        self,
        *,
        github_usernames: Union[set[str], None] = None,
        launchpad_usernames: Union[set[str], None] = None,
    ) -> set[str]:
        """
        Remove SSH keys that were imported from GitHub or Launchpad, and return the remaining set of
        keys.

        :type github_usernames: set[str] or None
        :param github_usernames:
            A set of GitHub usernames to remove SSH keys for (keyword-only argument)

        :type launchpad_usernames: set[str] or None
        :param launchpad_usernames:
            A set of Launchpad usernames to remove SSH keys for (keyword-only argument)

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the Pi is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        ssh_keys = self.ssh_keys
        if github_usernames:
            for username in github_usernames:
                ssh_keys = remove_imported_ssh_keys(ssh_keys, "gh", username)

        if launchpad_usernames:
            for username in launchpad_usernames:
                ssh_keys = remove_imported_ssh_keys(ssh_keys, "lp", username)

        self.ssh_keys = ssh_keys
        return self.ssh_keys

    def remove_ssh_keys(self, label: Union[str, None] = None) -> set[str]:
        """
        Remove an SSH key from the Pi that has a specific label (e.g. ``user@hostname``) and return
        the remaining set of keys. If *label* is ``None``, all keys will be removed.

        :type label: str or None
        :param label: The label of the SSH key to remove

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiProvisioningError:
            If the Pi is still provisioning

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        if label is None:
            self.ssh_keys = None
        else:
            self.ssh_keys = remove_ssh_keys_by_label(self.ssh_keys, label)
        return self.ssh_keys

    def wait_until_provisioned(self):
        """
        Wait for the new Pi to be provisioned

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        while True:
            pi_info = self.get_provision_status()
            if type(pi_info) is PiInfo:
                return
            sleep(5)

    def get_provision_status(self) -> Union[PiInfo, ProvisioningServer, None]:
        """
        Send a request to the server creation status endpoint and return the status as either a
        :class:`~hostedpi.models.mythic.responses.PiInfo` or
        :class:`~hostedpi.models.mythic.responses.ProvisioningServer` or ``None`` if the status is
        not yet available.

        :raises HostedPiNotAuthorizedError:
            If the user is not authorised to access the server

        :raises HostedPiServerError:
            If there is another error accessing the API
        """
        if self._status_url is None:
            return self.info

        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-queuepitask
        try:
            response = self.session.get(self._status_url)
        except ConnectionError as exc:
            logger.warn("Temporary error getting server provisioning status", exc=str(exc))
            return

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            raise HostedPiServerError(error) from exc

        log_request(response)

        status = self._parse_status(response.json())
        if type(status) is ProvisioningServer:
            logger.info("Server provisioning in progress", status=status.provision_status)
            return status
        if type(status) is PiInfo:
            self._name = response.request.url.split("/")[-1]
            logger.info("Server provisioning complete", server_name=self._name)
            self._info = status
            self._last_fetched_info = datetime.now(timezone.utc)
            self._status_url = None
            return status

    def _get_info(self):
        """
        Fetch the full Pi information from the API, or return immediately if the last fetch was
        less than 10 seconds ago.
        """
        if self.name is None:
            raise HostedPiUserError("Cannot fetch info for a Pi without a name")
        now = datetime.now(timezone.utc)
        if self._last_fetched_info is not None:
            if (now - self._last_fetched_info).total_seconds() < 10:
                return
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-piserversidentifier
        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}")
        response = self.session.get(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiProvisioningError(error) from exc
            raise HostedPiServerError(error) from exc

        self._info = PiInfo.model_validate(response.json())

    def _power_on_off(self, *, on: bool):
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-put-piserversidentifierpower
        url = urllib.parse.urljoin(self._api_url, f"servers/{self.name}/power")
        data = {
            "power": on,
        }
        response = self.session.put(url, json=data)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiProvisioningError(error) from exc
            raise HostedPiServerError(error) from exc

    def _parse_status(self, data: dict) -> Union[PiInfo, ProvisioningServer, None]:
        """
        Get the status of an async server creation request
        """
        try:
            return PiInfo.model_validate(data)
        except ValidationError:
            pass

        try:
            return ProvisioningServer.model_validate(data)
        except ValidationError:
            logger.warn("Unexpected response from server creation status endpoint")
            return



from pathlib import Path
from typing import Annotated, Union

from typer import Argument


server_name = Annotated[str, Argument(help="Name of the Raspberry Pi server")]
server_names = Annotated[Union[list[str], None], Argument(help="Names of the Raspberry Pi servers")]
ssh_key_path = Annotated[
    Path, Argument(help="Path to the SSH key to install on the Raspberry Pi servers")
]
images_model = Annotated[
    int, Argument(help="Raspberry Pi model number to list images for", min=3, max=4)
]
ssh_key_label = Annotated[str, Argument(help="Label for the SSH key, e.g. 'ben@finn'")]



import urllib.parse
from typing import Union

from pydantic import ValidationError
from requests import HTTPError, Session
from structlog import get_logger

from .auth import MythicAuth
from .exc import (
    HostedPiInvalidParametersError,
    HostedPiNameExistsError,
    HostedPiNotAuthorizedError,
    HostedPiOutOfStockError,
    HostedPiServerError,
    HostedPiUserError,
    HostedPiValidationError,
)
from .logger import log_request
from .models.mythic.payloads import NewServer
from .models.mythic.responses import (
    PiImagesResponse,
    PiInfoBasic,
    ServerSpec,
    ServersResponse,
    SpecsResponse,
)
from .models.specs import Pi3ServerSpec, Pi4ServerSpec
from .models.sshkeys import SSHKeySources
from .pi import Pi
from .utils import get_error_message


logger = get_logger()


class PiCloud:
    """
    A connection to the Mythic Beasts Pi Cloud API for creating and managing cloud Pi servers.

    :type ssh_keys: :class:`~hostedpi.models.sshkeys.SSHKeySources` or None
    :param ssh_keys:
        An instance of :class:`~hostedpi.models.sshkeys.SSHKeySources` containing sources of SSH
        keys to use when creating new Pis. If not provided, no SSH keys will be used by default.

    :type auth: :class:`~hostedpi.auth.MythicAuth` or None
    :param auth:
        An instance of :class:`~hostedpi.auth.MythicAuth` to use for authentication with the API.
        If not provided, a default instance will be created. You almost certainly won't need to
        set this yourself.

    .. note::
        If any SSH keys are provided on class initialisation, they will be used when creating Pis
        but are overriden by any passed to the :meth:`~hostedpi.picloud.PiCloud.create_pi` method.
    """

    def __init__(
        self,
        ssh_keys: Union[SSHKeySources, None] = None,
        *,
        auth: Union[MythicAuth, None] = None,
    ):
        self.ssh_keys = None
        if ssh_keys is not None:
            if not isinstance(ssh_keys, SSHKeySources):
                raise TypeError("ssh_keys must be an instance of SSHKeySources or None")
            self.ssh_keys = ssh_keys.collect()
        if auth is None:
            auth = MythicAuth()
        self._auth = auth
        self._api_url = str(auth.settings.api_url)

    def __repr__(self):
        return f"<PiCloud id={self._auth._settings.id}>"

    @property
    def session(self) -> Session:
        return self._auth.session

    @property
    def pis(self) -> dict[str, Pi]:
        """
        A dict of all Raspberry Pi servers associated with the account, keyed by their names.
        Each value is an instance of :class:`~hostedpi.pi.Pi` representing the server.

        :raises HostedPiNotAuthorizedError:
            If the user is not authorized to retrieve the list of Pis

        :raises HostedPiServerError:
            If there is an error retrieving the list from the server
        """
        servers = self._get_pis()
        return {
            name: Pi(name, info=info, auth=self._auth) for name, info in sorted(servers.items())
        }

    @property
    def ipv4_ssh_config(self) -> str:
        """
        A string containing the default IPv4 SSH config for all Pis within the account. The contents
        could be added to an SSH config file for easy access to the Pis in the account.
        """
        return self.get_ipv4_ssh_config()

    @property
    def ipv6_ssh_config(self) -> str:
        """
        A string containing the default IPv6 SSH config for all Pis within the account. The contents
        could be added to an SSH config file for easy access to the Pis in the account.
        """
        return self.get_ipv6_ssh_config()

    def get_ipv4_ssh_config(self, user: str = "root") -> str:
        """
        Construct a string containing the IPv4 SSH config for all Pis within the account. The
        contents could be added to an SSH config file for easy access to the Pis in the account.
        """
        return "\n".join(pi.get_ipv4_ssh_config(user=user) for pi in self.pis.values())

    def get_ipv6_ssh_config(self, user: str = "root", numeric: bool = False) -> str:
        """
        Construct a string containing the IPv6 SSH config for all Pis within the account. The
        contents could be added to an SSH config file for easy access to the Pis in the account.
        """
        return "\n".join(
            pi.get_ipv6_ssh_config(user=user, numeric=numeric) for pi in self.pis.values()
        )

    def create_pi(
        self,
        *,
        name: Union[str, None] = None,
        spec: Union[Pi3ServerSpec, Pi4ServerSpec],
        ssh_keys: Union[SSHKeySources, None] = None,
        wait: bool = False,
    ) -> Pi:
        """
        Provision a new cloud Pi with the specified name, model, disk size and SSH keys. Return a
        new :class:`~hostedpi.pi.Pi` instance.

        :type name: str or None
        :param name:
            A unique identifier for the server. This will form part of the hostname for the server,
            and must consist only of alphanumeric characters and hyphens. If not provided, a server
            name will be automatically generated.

        :type spec: Pi3ServerSpec or Pi4ServerSpec
        :param spec:
            The spec of the Raspberry Pi to provision

        :type ssh_keys: :class:`~hostedpi.models.sshkeys.SSHKeySources` or None
        :param ssh_keys:
            An instance of :class:`~hostedpi.models.sshkeys.SSHKeySources` containing sources of SSH
            keys to use when creating a new Pi. If not provided, no SSH keys will be added on
            creation.

        :type wait: bool
        :param wait:
            If True, the method will return immediately after the server creation request is
            accepted, without waiting for the server to be provisioned. The returned
            :class:`~hostedpi.pi.Pi` instance will not be fully initialised and will not be able to
            perform actions until the server is ready. If False, the method will wait for the server
            to be provisioned before returning the :class:`~hostedpi.pi.Pi` instance. Default is
            False.

        .. note::
            If any SSH keys are provided on class initialisation, they will be used here but are
            overriden by any passed to this method.

        .. note::
            When requesting a Pi 3, you will either get a model 3B or 3B+. It is not possible to
            request a particular model beyond 3 or 4. Some memory and CPU speed options are
            available when requesting a Pi 4.

        :raises HostedPiValidationError:
            If the provided name or spec is invalid

        :raises HostedPiInvalidParametersError:
            If the provided parameters are invalid, such as an unsupported model or disk size

        :raises HostedPiNotAuthorizedError:
            If the user is not authorized to create a new Pi server

        :raises HostedPiOutOfStockError:
            If there are no available Pi servers of the requested type

        :raises HostedPiServerError:
            If there is another error from the server
        """
        if name is None:
            # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-post-piservers
            url = urllib.parse.urljoin(self._api_url, "servers")
        else:
            # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-post-piserversidentifier
            url = urllib.parse.urljoin(self._api_url, f"servers/{name}")

        if not isinstance(spec, (Pi3ServerSpec, Pi4ServerSpec)):
            raise TypeError("spec must be an instance of Pi3ServerSpec or Pi4ServerSpec")

        if ssh_keys is not None:
            if not isinstance(ssh_keys, SSHKeySources):
                raise TypeError("ssh_keys must be an instance of SSHKeySources or None")
            ssh_keys = ssh_keys.collect()
        else:
            ssh_keys = self.ssh_keys

        try:
            data = NewServer(name=name, spec=spec, ssh_keys=ssh_keys)
        except ValidationError as exc:
            logger.error(f"Invalid server name or spec: {exc}")
            raise HostedPiValidationError("Invalid server name or spec") from exc

        num_ssh_keys = len(ssh_keys) if ssh_keys else 0
        logger.info("Creating new server", name=name, spec=spec, ssh_keys=num_ssh_keys)
        response = self.session.post(url, json=data.payload)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 400:
                raise HostedPiInvalidParametersError(error) from exc
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            if response.status_code == 409:
                raise HostedPiNameExistsError(error) from exc
            if response.status_code == 503:
                raise HostedPiOutOfStockError(error) from exc
            raise HostedPiServerError(error) from exc

        status_url = response.headers["Location"]

        logger.info("Server creation request accepted", status_url=status_url)
        basic_info = PiInfoBasic.model_validate(spec)
        pi = Pi(
            name=name,
            info=basic_info,
            auth=self._auth,
            status_url=status_url,
        )
        if wait:
            pi.wait_until_provisioned()
        return pi

    def get_operating_systems(self, *, model: int) -> dict[str, str]:
        """
        Return a dict of operating systems supported by the given Pi *model* (3 or 4). Dict keys are
        identifiers (e.g. "rpi-bookworm-armhf") which can be used when provisioning a new Pi with
        :meth:`~hostedpi.picloud.PiCloud.create_pi`; dict values are text labels of the OS/distro
        names (e.g. "Raspberry Pi OS Bookworm (32 bit)").

        :type model: int
        :param model:
            The Raspberry Pi model (3 or 4) to get operating systems for (keyword-only argument)

        :raises HostedPiUserError:
            If the provided model is not 3 or 4

        :raises HostedPiServerError:
            If there is an error retrieving the operating systems from the server
        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-piimagesmodel
        if model not in {3, 4}:
            raise HostedPiUserError("model must be 3 or 4")
        url = urllib.parse.urljoin(self._api_url, f"images/{model}")
        response = self.session.get(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            raise HostedPiServerError(error) from exc

        return PiImagesResponse.model_validate(response.json()).root

    def _get_available_specs(self) -> list[ServerSpec]:
        """
        Retrieve all available Raspberry Pi server specifications

        :raises HostedPiServerError:
            If there is an error retrieving the specifications from the server
        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-pimodels
        url = urllib.parse.urljoin(self._api_url, "models")
        response = self.session.get(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            raise HostedPiServerError(error) from exc

        data = SpecsResponse.model_validate(response.json())
        return data.models

    def _get_pis(self) -> dict[str, PiInfoBasic]:
        """
        Retrieve all Raspberry Pi servers associated with the account
        """
        # https://www.mythic-beasts.com/support/api/raspberry-pi#ep-get-piservers
        url = urllib.parse.urljoin(self._api_url, "servers")

        response = self.session.get(url)
        log_request(response)

        try:
            response.raise_for_status()
        except HTTPError as exc:
            error = get_error_message(exc)
            if response.status_code == 403:
                raise HostedPiNotAuthorizedError(error) from exc
            raise HostedPiServerError(error) from exc

        response = ServersResponse.model_validate(response.json())
        return response.servers



from ipaddress import IPv6Address, IPv6Network
from typing import Union

from pydantic import BaseModel, ConfigDict, Field, RootModel, model_validator


class AuthResponse(BaseModel):
    """
    Response from the Mythic Beasts Authentication API
    """

    access_token: str
    expires_in: int


class PiInfoBasic(BaseModel):
    """
    Basic information about a Raspberry Pi server, which comes back when listing servers
    """

    model_config = ConfigDict(from_attributes=True)

    model: int = Field(description="The model number of the Raspberry Pi server")
    memory: int = Field(description="The Pi's RAM size in MB")
    cpu_speed: int = Field(description="The Pi's CPU speed in MHz")

    @property
    def memory_gb(self) -> int:
        """
        The Pi's RAM size in GB
        """
        return self.memory // 1024


class ServersResponse(BaseModel):
    """
    Response from the Mythic Beasts API when listing servers
    """

    servers: dict[str, PiInfoBasic]


class ErrorResponse(BaseModel):
    """
    Response from the Mythic Beasts API when an API error occurs
    """

    error: str = "Error"


class ProvisioningServer(BaseModel):
    """
    Response from the Mythic Beasts API when a Raspberry Pi server is still provisioning
    """

    provision_status: str = Field(
        alias="status", description="The provisioning status of the server"
    )


class PiInfo(PiInfoBasic, ProvisioningServer):
    """
    Detailed information about a Raspberry Pi server, which comes back when retrieving details for a
    specific server
    """

    model_full: Union[str, None] = Field(
        default=None,
        description="The full model name of the server, e.g. 3B, 3B+ or 4B rather than just the model number",
    )
    is_booting: bool = Field(description="Whether the server is currently booting")
    boot_progress: Union[str, None] = Field(
        default=None, description="The server's current boot progress"
    )
    power: bool = Field(
        description="The server's power state, True if powered on, False if powered off"
    )
    ssh_port: int = Field(
        description="The SSH port that can be used to connect to the server using the IPv4 proxy"
    )
    disk_size: Union[int, None] = Field(default=None, description="The disk size in GB")
    nic_speed: int = Field(description="The network interface speed of the server in Mbps")
    ipv6_address: IPv6Address = Field(alias="ip", description="The IPv6 address of the server")
    ipv6_network: IPv6Network = Field(
        alias="ip_routed", description="The IPv6 network the server is on"
    )
    initialised_keys: bool = Field(description="Whether the server has initialised SSH keys")
    location: str = Field(description="The location of the data centre the server is in")


class SSHKeysResponse(BaseModel):
    """
    Response from the Mythic Beasts API when retrieving SSH keys
    """

    keys_raw: str = Field(alias="ssh_key", default="")
    keys: set[str] = set()

    @model_validator(mode="after")
    def make_keys_set(self):
        self.keys = {key.strip() for key in self.keys_raw.split("\n") if key.strip()}
        return self


class PiImagesResponse(RootModel):
    """
    Response from the Mythic Beasts API when retrieving available Pi images
    """

    root: dict[str, str]


class ServerSpec(BaseModel):
    """
    Details for a server specification, which comes back when retrieving available server
    specifications
    """

    disk: int = 10
    model: int
    memory: int
    cpu_speed: int
    nic_speed: int

    @property
    def memory_gb(self) -> int:
        return self.memory // 1024


class SpecsResponse(BaseModel):
    """
    Response from the Mythic Beasts API when retrieving available server specifications
    """

    models: list[ServerSpec]



from pydantic import AnyHttpUrl, Field, SecretStr, field_validator
from pydantic_settings import BaseSettings, SettingsConfigDict


class Settings(BaseSettings):
    """
    This model handles the configuration settings for the Mythic Beasts Hosted Pi API.

    :type id: str
    :param id:
        Your API ID. Required to authenticate with the API.

    :type secret: str
    :param secret:
        Your API secret key. Required to authenticate with the API.

    :type auth_url: str
    :param auth_url:
        The authentication URL for the API. This is used to log in and obtain an access token.
        Defaults to "https://auth.mythic-beasts.com/login".

    :type api_url: str
    :param api_url:
        The base URL for the API. This is used to make requests to the API endpoints.
        Defaults to "https://api.mythic-beasts.com/beta/pi/".

    :raises pydantic_core.ValidationError:
        If the provided settings are invalid or missing required fields.
    """

    model_config = SettingsConfigDict(env_prefix="hostedpi_", env_file=".env", extra="ignore")

    id: str = Field(description="Your API ID")
    secret: SecretStr = Field(description="Your API secret key")
    auth_url: AnyHttpUrl = Field(
        default="https://auth.mythic-beasts.com/login",
        description="The authentication URL",
    )
    api_url: AnyHttpUrl = Field(
        default="https://api.mythic-beasts.com/beta/pi/",
        description="The API URL",
    )

    @field_validator("api_url", mode="before")
    @classmethod
    def ensure_trailing_slash(cls, v):
        if not v.endswith("/"):
            v += "/"
        return v



import string
from typing import Union

from pydantic import BaseModel, field_validator

from ..specs import Pi3ServerSpec, Pi4ServerSpec


class NewServer(BaseModel):
    name: Union[str, None] = None
    spec: Union[Pi3ServerSpec, Pi4ServerSpec]
    ssh_keys: Union[set[str], None] = None

    @field_validator("name", mode="after")
    @classmethod
    def validate_name(cls, v):
        if v is None or v == "":
            return None
        server_name = v.lower()
        valid_chars = string.ascii_lowercase + string.digits + "-"
        if not all(c in valid_chars for c in server_name):
            raise ValueError("Server name must consist of alphanumeric characters and hyphens")
        return server_name

    @field_validator("ssh_keys", mode="before")
    @classmethod
    def validate_ssh_keys(cls, v):
        if v == set():
            return None
        return v

    @property
    def payload(self) -> dict:
        data = self.spec.model_dump(exclude_none=True)
        if self.ssh_keys is not None:
            data["ssh_key"] = "\r\n".join(self.ssh_keys)
        return data



class HostedPiException(Exception):
    "Base class for all exceptions in the hostedpi module"


class MythicAuthenticationError(HostedPiException):
    "Exception raised when authentication fails"


class HostedPiUserError(HostedPiException):
    "Base exception for user errors, such as invalid input or configuration issues"


class HostedPiValidationError(HostedPiUserError):
    "Exception raised for validation errors, such as invalid data formats or missing required fields"


class HostedPiServerError(HostedPiException):
    "Exception raised for server errors, such as API errors or server-side issues"


class HostedPiInvalidParametersError(HostedPiServerError, HostedPiUserError):
    "Exception raised for invalid parameters in API requests or server operations"


class HostedPiNotAuthorizedError(HostedPiServerError):
    "Exception raised when the user is not authorized to perform a specific action or access a resource"


class HostedPiOutOfStockError(HostedPiServerError):
    "Exception raised when no servers with the required specification are available"


class HostedPiProvisioningError(HostedPiServerError):
    "Exception raised when an action cannot be performed because the Pi is still being provisioned"


class HostedPiNameExistsError(HostedPiServerError):
    "Exception raised when a Pi with the specified name already exists"



from pathlib import Path
from typing import Literal, Union

import requests
from requests.exceptions import HTTPError
from structlog import get_logger

from .logger import log_request
from .models.mythic.responses import ErrorResponse


logger = get_logger()


def ssh_import_id(
    *,
    github_username: Union[str, None] = None,
    launchpad_username: Union[str, None] = None,
) -> set[str]:
    """
    Returns a set of SSH keys imported from GitHub and/or Launchpad
    """
    keys = set()
    if github_username is not None:
        url = f"https://github.com/{github_username}.keys"
        sep = "\n"
        keys |= {
            _add_ssh_import_tag(key, "gh", github_username) for key in fetch_keys_from_url(url, sep)
        }
    if launchpad_username is not None:
        url = f"https://launchpad.net/~{launchpad_username}/+sshkeys"
        sep = "\r\n\n"
        keys |= {
            _add_ssh_import_tag(key, "lp", launchpad_username)
            for key in fetch_keys_from_url(url, sep)
        }

    return keys


def fetch_keys_from_url(url: str, sep: str) -> set[str]:
    """
    Retrieve keys from *url* and return a set of keys
    """
    response = requests.get(url)
    log_request(response)
    response.raise_for_status()
    return set(response.text.strip().split(sep))


def collect_ssh_keys(
    *,
    ssh_keys: Union[set[str], None] = None,
    ssh_key_path: Union[Path, None] = None,
    github_usernames: Union[set[str], None] = None,
    launchpad_usernames: Union[set[str], None] = None,
) -> set[str]:
    """
    Collect and combine SSH keys from any of various sources
    """
    ssh_keys_set = set()
    if ssh_keys:
        ssh_keys_set |= ssh_keys
    if ssh_key_path:
        ssh_keys_set |= {ssh_key_path.read_text().strip()}
    if github_usernames:
        for username in github_usernames:
            ssh_keys_set |= ssh_import_id(github_username=username)
    if launchpad_usernames:
        for username in launchpad_usernames:
            ssh_keys_set |= ssh_import_id(launchpad_username=username)
    return dedupe_ssh_keys(ssh_keys_set)


def dedupe_ssh_keys(ssh_keys: set[str]) -> set[str]:
    """
    Deduplicate SSH keys by removing any duplicates that are identical except for the comment at the
    end of the key
    """

    def keysort(key: str) -> int:
        return (-key.count(" "), key)

    # sort keys by the number of spaces so we don't throw away any additional comments
    sorted_keys = sorted(ssh_keys, key=keysort)

    actual_keys = set()  # without comments
    unique_keys = set()  # with comments (returned)

    for key in sorted_keys:
        # Split the key into the actual key and the comment
        parts = key.split()
        actual_key = " ".join(parts[:2])
        if actual_key not in actual_keys:
            # Only add unique keys to the return set
            unique_keys.add(key)
            actual_keys.add(actual_key)

    return unique_keys


def get_error_message(exc: HTTPError) -> str:
    """
    Try to retrieve an error message from the response
    """
    try:
        data = exc.response.json()
    except Exception:
        if exc.response.text:
            return f"Error {exc.response.status_code}: {exc.response.text}"
        return f"Error {exc.response.status_code}"

    return ErrorResponse.model_validate(data).error


def remove_ssh_keys_by_label(ssh_keys: set[str], label: str) -> set[str]:
    """
    Remove SSH keys that have a specific label (e.g. ``user@hostname``)
    """
    if label:
        return {key for key in ssh_keys if _extract_ssh_key_label(key) != label}
    return ssh_keys


def remove_imported_ssh_keys(
    ssh_keys: set[str], source: Literal["gh", "lp"], username: str
) -> set[str]:
    """
    Remove SSH keys that were imported from a specific source (GitHub or Launchpad)
    """
    return {key for key in ssh_keys if not _is_imported_ssh_key(key, source, username)}


def _extract_ssh_key_label(key: str) -> Union[str, None]:
    """
    Try to extract the label (e.g. ``user@hostname``) from an SSH key string, otherwise return None
    """
    parts = key.split(" ")
    if len(parts) > 2 and "@" in parts[2]:
        return parts[2]


def _is_imported_ssh_key(key: str, source: Literal["gh", "lp"], username: str) -> bool:
    """
    Check if the SSH key was imported from a specific source (GitHub or Launchpad) and username
    """
    import_comment = f"# ssh-import-id {source}:{username}"
    return key.endswith(import_comment)


def _add_ssh_import_tag(key: str, source: str, username: str) -> str:
    """
    Add a tag to the SSH key to indicate it was imported
    """
    return f"{key} # ssh-import-id {source}:{username}"



from typing import Union


def memory(v: Union[int, None]) -> str:
    """
    Format memory size to GB
    """
    if v is None:
        return ""
    return f"{v} GB"


def cpu_speed(v: Union[int, None]) -> str:
    """
    Format CPU speed to GHz
    """
    if v is None:
        return ""
    return f"{v / 1000:.1f} GHz"


def nic_speed(v: Union[int, None]) -> str:
    """
    Format NIC speed to Mbps or Gbps
    """
    return {
        100: "100 Mbps",
        1000: "1 Gbps",
    }.get(v, "")


def disk_size(v: Union[int, None]) -> str:
    """
    Format disk size to GB or TB
    """
    if v is None:
        return ""
    if v < 1024:
        return f"{v} GB"
    return f"{v / 1024:.1f} TB"


def boolean(v: bool) -> str:
    """
    Format boolean to Yes or No
    """
    return "Yes" if v else "No"



from typing import Union

from pydantic import BaseModel, Field, FilePath

from ..utils import collect_ssh_keys


class SSHKeySources(BaseModel):
    """
    Sources for SSH keys to be added to Pi servers

    :type ssh_keys: set[str] | None
    :param ssh_keys: Set of SSH key strings

    :type ssh_key_path: :class:`~pathlib.Path` | None
    :param ssh_key_path: Path to an SSH key file

    :type github_usernames: set[str] | None
    :param github_usernames: Set of GitHub usernames to collect SSH keys for

    :type launchpad_usernames: set[str] | None
    :param launchpad_usernames: Set of Launchpad usernames to collect SSH keys for

    :raises pydantic_core.ValidationError:
        If the SSH key sources are invalid
    """

    ssh_keys: Union[set[str], None] = Field(default=None, description="Set of SSH key strings")
    ssh_key_path: Union[FilePath, None] = Field(default=None, description="Path to an SSH key file")
    github_usernames: Union[set[str], None] = Field(
        default=None, description="Set of GitHub usernames to collect SSH keys for"
    )
    launchpad_usernames: Union[set[str], None] = Field(
        default=None, description="Set of Launchpad usernames to collect SSH keys for"
    )

    def collect(self) -> Union[set[str], None]:
        """
        Collect SSH keys from various sources, and return them as a set of strings which can be
        added to a Pi by setting :attr:`~hostedpi.pi.Pi.ssh_keys`.
        """
        keys = collect_ssh_keys(
            ssh_keys=self.ssh_keys,
            ssh_key_path=self.ssh_key_path,
            github_usernames=self.github_usernames,
            launchpad_usernames=self.launchpad_usernames,
        )
        return keys if keys else None



import rich
from rich.console import Console
from rich.live import Live
from typer import Exit, Typer

from ..exc import HostedPiException
from . import arguments, options, utils
from .ssh import ssh_app


app = Typer(name="hostedpi", no_args_is_help=True)
app.add_typer(ssh_app, name="ssh", no_args_is_help=True, help="SSH access management commands")
console = Console()


@app.command("connect", hidden=True)
@app.command("test", help="Test a connection to the Mythic Beasts API")
def do_test():
    try:
        cloud = utils.get_picloud()
        cloud._auth.token
        utils.print_success("Connected to the Mythic Beasts API")
    except Exception:
        utils.print_error("Failed to authenticate")
        raise Exit(1)


@app.command("images")
def do_images(
    model: arguments.images_model,
    filter: options.filter_pattern_images = None,
):
    """
    List operating system images available for Raspberry Pi servers
    """
    cloud = utils.get_picloud()
    images = cloud.get_operating_systems(model=model)
    table = utils.make_table("ID", "Name")

    for id, name in sorted(images.items()):
        if filter is None or filter.lower() in id.lower() or filter.lower() in name.lower():
            table.add_row(id, name)
    rich.print(table)


@app.command("ls", hidden=True)
@app.command("list")
def do_list(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
):
    """
    List Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)

    for pi in pis:
        print(pi.name)


@app.command("tab", hidden=True)
@app.command("table")
def do_table(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
    full: options.full_table = False,
):
    """
    List Raspberry Pi server information in a table
    """
    pis = utils.get_pis(names, filter)

    if full:
        utils.full_pis_table(pis)
    else:
        utils.short_pis_table(pis)


@app.command("create")
def do_create(
    model: options.model,
    names: arguments.server_names = None,
    number: options.number = None,
    disk: options.disk_size = 10,
    memory: options.memory = None,
    cpu_speed: options.cpu_speed = None,
    os_image: options.os_image = None,
    wait: options.wait = False,
    ssh_key_path: options.ssh_key_path = None,
    ssh_import_github: options.ssh_import_github = None,
    ssh_import_launchpad: options.ssh_import_launchpad = None,
    full: options.full_table = False,
):
    """
    Provision one or more new Raspberry Pi servers
    """
    if names and number:
        utils.print_error("You can't specify both names and a number")
        raise Exit(1)
    if not names and not number:
        number = 1
    if full and not wait:
        utils.print_error("You can't use --full without --wait")
        raise Exit(1)
    if ssh_key_path is not None:
        if not ssh_key_path.exists():
            utils.print_error(f"SSH key file not found: {ssh_key_path}")
            raise Exit(1)
    ssh_import_gh_set = set(ssh_import_github) if ssh_import_github is not None else None
    ssh_import_lp_set = set(ssh_import_launchpad) if ssh_import_launchpad is not None else None

    if names:
        for name in names:
            try:
                utils.create_pi(
                    name=name,
                    model=model,
                    disk=disk,
                    memory_gb=memory,
                    cpu_speed=cpu_speed,
                    os_image=os_image,
                    wait=wait,
                    ssh_key_path=ssh_key_path,
                    github_usernames=ssh_import_gh_set,
                    launchpad_usernames=ssh_import_lp_set,
                    full=full,
                )
            except HostedPiException as exc:
                utils.print_exc(exc)
                continue

    if number:
        for n in range(number):
            try:
                utils.create_pi(
                    model=model,
                    disk=disk,
                    memory_gb=memory,
                    cpu_speed=cpu_speed,
                    os_image=os_image,
                    wait=wait,
                    ssh_key_path=ssh_key_path,
                    github_usernames=ssh_import_gh_set,
                    launchpad_usernames=ssh_import_lp_set,
                    full=full,
                )
            except HostedPiException as exc:
                utils.print_exc(exc)
                continue


@app.command("status")
def do_status(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Get the current status of one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    table = utils.make_table("Name", "Status")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                table.add_row(pi.name, pi.status)
            except HostedPiException as exc:
                utils.print_exc(exc)
                table.add_row(pi.name, "Error")
                continue


@app.command("on")
def do_on(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Power on one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    table = utils.make_table("Name", "Status")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                pi.on()
            except HostedPiException as exc:
                utils.print_exc(exc)
                table.add_row(pi.name, "Error")
                continue
            table.add_row(pi.name, "Powering on")


@app.command("off")
def do_off(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Power off one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    table = utils.make_table("Name", "Status")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                pi.off()
            except HostedPiException as exc:
                utils.print_exc(exc)
                table.add_row(pi.name, "Error")
                continue
            table.add_row(pi.name, "Powering off")


@app.command("reboot")
def do_reboot(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Reboot one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    table = utils.make_table("Name", "Status")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                pi.reboot()
            except HostedPiException as exc:
                utils.print_exc(exc)
                continue
            table.add_row(pi.name, "Rebooting")


@app.command("rm", hidden=True)
@app.command("cancel")
def do_cancel(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
    yes: options.yes = False,
):
    """
    Unprovision one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    pis_str = ", ".join([pis.name for pis in pis])
    if len(pis) == 0:
        utils.print_error("No servers to cancel")
        raise Exit(1)
    if not yes:
        yn = input(f"Are you sure you want to cancel {pis_str}? [y/N] ")
        if yn.lower() != "y":
            raise Exit(1)
    table = utils.make_table("Name", "Status")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                pi.cancel()
            except HostedPiException as exc:
                table.add_row(pi.name, "Failed to cancel")
                utils.print_exc(exc)
                continue
            table.add_row(pi.name, "Cancelled")



from typing import Literal, Union

from pydantic import BaseModel, Field, field_validator, model_validator


class NewServerSpec(BaseModel):
    disk: int = Field(default=10, description="Disk size in GB")
    model: int
    memory: int = 0
    memory_gb: Union[int, None] = Field(default=None, exclude=True)
    cpu_speed: int
    os_image: Union[str, None] = Field(default=None, description="Operating system image")
    # wait_for_dns: bool = False

    @field_validator("disk", mode="after")
    @classmethod
    def validate_disk(cls, v):
        if v < 10:
            raise ValueError("Disk size must be at least 10GB")
        if v % 10 != 0:
            raise ValueError("Disk size must be a multiple of 10GB")
        return v

    @model_validator(mode="after")
    def convert_memory_gb_to_mb(self):
        if self.memory_gb is not None:
            self.memory = self.memory_gb * 1024
        return self


class Pi3ServerSpec(NewServerSpec):
    """
    Specification model for the Raspberry Pi 3

    :type disk: int
    :param disk: Disk size in GB. Must be a multiple of 10, defaults to 10.

    :type os_image: str | None
    :param os_image:
        Operating system image to use. Defaults to None, which uses Mythic's default image.

    :raises pydantic_core.ValidationError:
        If the server specification is invalid
    """

    model: Literal[3] = 3
    memory: Union[Literal[1024], None] = None
    memory_gb: Literal[3] = Field(
        default=1,
        exclude=True,
        description="Memory in GB. Pi 3 only supports 1GB.",
    )
    cpu_speed: Literal[1200] = Field(
        default=1200,
        description="CPU speed in MHz. Pi 3 only supports 1200MHz.",
    )


class Pi4ServerSpec(NewServerSpec):
    """
    Specification model for the Raspberry Pi 4

    :type disk: int
    :param disk: Disk size in GB. Must be a multiple of 10, defaults to 10.

    :type memory_gb: int
    :param memory_gb: Memory in GB. Can be 4 or 8, defaults to 4.

    :type cpu_speed: int
    :param cpu_speed: CPU speed in MHz. Can be 1500 or 2000, defaults to 1500.

    :type os_image: str | None
    :param os_image:
        Operating system image to use. Defaults to None, which uses Mythic's default image.

    :raises pydantic_core.ValidationError:
        If the server specification is invalid
    """

    model: Literal[4] = 4
    memory: Union[Literal[4096, 8192], None] = None
    memory_gb: Literal[4, 8] = Field(
        default=4,
        exclude=True,
        description="Memory in GB. Pi 4 supports 4GB or 8GB",
    )
    cpu_speed: Literal[1500, 2000] = Field(
        default=1500,
        description="CPU speed in MHz. Pi 4 supports 1500MHz or 2000MHz.",
    )

    @model_validator(mode="after")
    def validate_memory_cpu_combo(self):
        if self.memory_gb == 4 and self.cpu_speed == 2000:
            raise ValueError("4GB Pi 4 only supports 1500MHz CPU speed")
        return self



import rich
from rich.console import Console
from rich.live import Live
from rich.table import Table
from typer import Exit, Typer

from ..exc import HostedPiException
from ..models.sshkeys import SSHKeySources
from ..utils import remove_imported_ssh_keys, remove_ssh_keys_by_label
from . import arguments, options, utils


keys_app = Typer()
console = Console()


@keys_app.command("count")
def do_count(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Count the number of SSH keys on one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    table = utils.make_table("Name", "Keys")
    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            try:
                n = len(pi.ssh_keys)
            except HostedPiException as exc:
                utils.print_exc(exc)
                continue
            table.add_row(pi.name, str(n))


@keys_app.command("cat", hidden=True)
@keys_app.command("show")
def do_show(name: arguments.server_name):
    """
    List the SSH keys on a Raspberry Pi server as plaintext
    """
    pi = utils.get_pi(name)
    if pi is None:
        utils.print_error(f"Pi '{name}' not found")
        raise Exit(1)
    for key in pi.ssh_keys:
        try:
            print(key)
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue


@keys_app.command("list")
def do_list(name: arguments.server_name):
    """
    List the SSH keys on a Raspberry Pi server, using the key label and note if available
    """
    pi = utils.get_pi(name)
    if pi is None:
        utils.print_error(f"Pi '{name}' not found")
        raise Exit(1)
    keys = pi.ssh_keys
    if not keys:
        utils.print_warn(f"No SSH keys found on {pi.name}")
        return
    for key in keys:
        try:
            parts = key.split()
            print(" ".join(parts[2:]))
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue


@keys_app.command("table")
def do_table(name: arguments.server_name, filter: options.filter_pattern_pi = None):
    """
    List the SSH keys on a Raspberry Pi server in a table format
    """
    pi = utils.get_pi(name)
    if pi is None:
        utils.print_error(f"Pi '{name}' not found")
        raise Exit(1)
    headers = ["Type", "Label", "Note"]
    table = Table(*headers)

    for key in pi.ssh_keys:
        if filter and not filter.lower() in key.lower():
            continue
        label = ""
        note = ""
        parts = key.split(" ")
        key_type = parts[0]
        if len(parts) > 1:
            if "@" in parts[2]:
                label = parts[2]
                note = " ".join(parts[3:]).removeprefix("# ")
            else:
                note = " ".join(parts[2:]).removeprefix("# ")

        table.add_row(key_type, label, note)

    rich.print(table)


@keys_app.command("add")
def do_add(
    ssh_key_path: arguments.ssh_key_path,
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
):
    """
    Add an SSH key to one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    ssh_keys = SSHKeySources(ssh_key_path=ssh_key_path)
    for pi in pis:
        keys_before = len(pi.ssh_keys)
        try:
            pi.add_ssh_keys(ssh_keys)
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue
        keys_after = len(pi.ssh_keys)
        if keys_after == keys_before:
            utils.print_warn(f"Key {ssh_key_path} already exists on {pi.name}")
        else:
            utils.print_success(f"Added key {ssh_key_path} to {pi.name}")


@keys_app.command("cp", hidden=True)
@keys_app.command("copy")
def do_copy(src: arguments.server_name, dests: arguments.server_names):
    """
    Copy the SSH keys from one Raspberry Pi server to others
    """
    src_pi = utils.get_pi(src)
    dest_pis = utils.get_pis(dests)
    ssh_keys = src_pi.ssh_keys

    for dest_pi in dest_pis:
        num_keys_before = len(dest_pi.ssh_keys)

        try:
            dest_pi.ssh_keys |= ssh_keys
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue

        num_keys_after = len(dest_pi.ssh_keys)
        num_keys_copied = num_keys_after - num_keys_before
        if num_keys_copied == 0:
            utils.print_warn(f"No new keys copied to {dest_pi.name} from {src_pi.name}")
        elif num_keys_copied == 1:
            utils.print_success(f"Copied 1 key from {src_pi.name} to {dest_pi.name}")
        else:
            utils.print_success(
                f"Copied {num_keys_copied} keys from {src_pi.name} to {dest_pi.name}"
            )


@keys_app.command("rm", hidden=True)
@keys_app.command("remove")
def do_remove(
    label: arguments.ssh_key_label,
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
):
    """
    Remove an SSH key from one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    for pi in pis:
        keys = pi.ssh_keys
        keys_before = len(keys)
        try:
            keys = remove_ssh_keys_by_label(keys, label)
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue
        pi.ssh_keys = keys
        removed_keys = keys_before - len(keys)
        if removed_keys == 0:
            utils.print_warn(f"No keys matching '{label}' found on {pi.name}")
        else:
            utils.print_success(f"Removed '{label}' key from {pi.name}")


@keys_app.command("purge")
def do_purge(names: arguments.server_names = None, filter: options.filter_pattern_pi = None):
    """
    Remove all SSH keys from one or more Raspberry Pi servers
    """
    pis = utils.get_pis(names, filter)
    for pi in pis:
        keys = len(pi.ssh_keys)
        try:
            pi.ssh_keys = None
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue
        if keys == 0:
            utils.print_warn(f"No keys to remove from {pi.name}")
        elif keys == 1:
            utils.print_success(f"Removed 1 key from {pi.name}")
        else:
            utils.print_success(f"Removed {keys} keys from {pi.name}")


@keys_app.command("import")
def do_import(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
    github: options.ssh_import_github = None,
    launchpad: options.ssh_import_launchpad = None,
):
    """
    Import SSH keys from GitHub and/or Launchpad to one or more Raspberry Pi servers
    """
    if not github and not launchpad:
        utils.print_error("You must specify at least one source to import from")
        raise Exit(1)
    pis = utils.get_pis(names, filter)
    ssh_keys = SSHKeySources(
        github_usernames=set(github) if github else None,
        launchpad_usernames=set(launchpad) if launchpad else None,
    )
    for pi in pis:
        keys_before = len(pi.ssh_keys)
        try:
            pi.add_ssh_keys(ssh_keys)
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue
        keys_after = len(pi.ssh_keys)
        keys_imported = keys_after - keys_before
        if keys_imported == 0:
            utils.print_warn(f"No new keys imported to {pi.name}")
        elif keys_imported == 1:
            utils.print_success(f"Imported 1 key to {pi.name}")
        else:
            utils.print_success(f"Imported {keys_imported} keys to {pi.name}")


@keys_app.command("unimport")
def do_unimport(
    names: arguments.server_names = None,
    filter: options.filter_pattern_pi = None,
    github: options.ssh_import_github = None,
    launchpad: options.ssh_import_launchpad = None,
):
    """
    Remove imported SSH keys from one or more Raspberry Pi servers
    """
    if not github and not launchpad:
        utils.print_error("You must specify at least one source to unimport from")
        raise Exit(1)
    github = set(github) if github else set()
    launchpad = set(launchpad) if launchpad else set()
    pis = utils.get_pis(names, filter)
    for pi in pis:
        keys = pi.ssh_keys
        keys_before = len(keys)
        for gh_username in github:
            keys = remove_imported_ssh_keys(keys, "gh", gh_username)
        for lp_username in launchpad:
            keys = remove_imported_ssh_keys(keys, "lp", lp_username)
        try:
            pi.ssh_keys = keys
        except HostedPiException as exc:
            utils.print_exc(exc)
            continue
        removed_keys = keys_before - len(keys)
        if removed_keys == 0:
            utils.print_warn(f"No keys matching import sources specified found on {pi.name}")
        elif removed_keys == 1:
            utils.print_success(f"Removed 1 key from {pi.name}")
        else:
            utils.print_success(f"Removed {removed_keys} keys from {pi.name}")



from . import logger
from .auth import MythicAuth
from .models import Pi3ServerSpec, Pi4ServerSpec, PiInfo, SSHKeySources
from .pi import Pi
from .picloud import PiCloud
from .settings import Settings


__all__ = [
    "MythicAuth",
    "Pi",
    "PiCloud",
    "Pi3ServerSpec",
    "Pi4ServerSpec",
    "PiInfo",
    "SSHKeySources",
    "Settings",
]



import json
import os

import structlog
from requests import Response
from structlog import get_logger


log_level = os.getenv("HOSTEDPI_LOG_LEVEL", "ERROR")
structlog.configure(wrapper_class=structlog.make_filtering_bound_logger(log_level))
logger = get_logger()


def log_request(response: Response):
    """
    Log the request and response
    """
    logger.debug(
        "Sent request",
        url=response.url,
        method=response.request.method,
        status_code=response.status_code,
        body=json.loads(_get_response_body(response)) if response.request.body else "",
    )

    try:
        response_body = response.json()
    except Exception:
        response_body = response.text
    logger.debug("Received response", body=response_body)


def _get_response_body(response: Response) -> str:
    if response.request.body:
        if isinstance(response.request.body, bytes):
            return response.request.body.decode("utf-8")
        return response.request.body
    return ""



from datetime import datetime, timedelta
from importlib.metadata import version
from typing import Union

from pydantic import ValidationError
from requests import HTTPError, Session
from structlog import get_logger

from .exc import MythicAuthenticationError
from .models.mythic.responses import AuthResponse
from .settings import Settings


hostedpi_version = version("hostedpi")
logger = get_logger()


class MythicAuth:
    """
    This class handles authentication with the Mythic Beasts Hosted Pi API.

    It manages the access token used to authenticate requests to the API, and automatically
    refreshes it when it expires.

    :type settings: :class:`~hostedpi.settings.Settings` or None
    :param settings:
        The settings used to configure the API. If not provided, defaults to a new
        instance of :class:`~hostedpi.settings.Settings`.

    :type auth_session: :class:`requests.Session` or None
    :param auth_session:
        The session used to make authentication requests. If not provided, defaults to a new
        :class:`requests.Session`.

    :type api_session: :class:`requests.Session` or None
    :param api_session:
        The session used to make API requests. If not provided, defaults to a new
        :class:`requests.Session`.

    .. warning::
        This is for advanced use only. Most users should not need to interact with the
        authentication system directly, as it is handled automatically by
        :class:`~hostedpi.picloud.PiCloud`.

    :raises pydantic_core.ValidationError:
        If the provided settings are invalid or missing required fields.
    """

    def __init__(
        self,
        *,
        settings: Union[Settings, None] = None,
        auth_session: Union[Session, None] = None,
        api_session: Union[Session, None] = None,
    ):
        if settings is None:
            settings = Settings()
        if auth_session is None:
            auth_session = Session()
        if api_session is None:
            api_session = Session()
        self._settings = settings
        self._token = None
        self._token_expiry = datetime.now()
        api_session.headers = {
            "User-Agent": f"python-hostedpi/{hostedpi_version}",
        }
        self._auth_session = auth_session
        self._api_session = api_session

    def __repr__(self):
        return f"<MythicAuth id={self._settings.id}>"

    @property
    def session(self) -> Session:
        """
        The session used to make requests to the Hosted Pi API
        """
        self._api_session.headers["Authorization"] = f"Bearer {self.token}"
        return self._api_session

    @property
    def settings(self) -> Settings:
        """
        The settings used to configure the API
        """
        return self._settings

    @property
    def token(self) -> str:
        """
        The access token used to authenticate requests to the API. The token is automatically
        refreshed when it expires.
        """
        if self._token is None or datetime.now() > self._token_expiry:
            data = {"grant_type": "client_credentials"}
            creds = (self.settings.id, self.settings.secret.get_secret_value())
            logger.debug("Authenticating", client_id=self.settings.id)
            response = self._auth_session.post(str(self.settings.auth_url), auth=creds, data=data)

            try:
                response.raise_for_status()
            except HTTPError as exc:
                logger.debug("Failed to authenticate", error=str(exc))
                raise MythicAuthenticationError("Failed to authenticate") from exc

            try:
                body = AuthResponse.model_validate(response.json())
            except ValidationError as exc:
                logger.debug("Failed to validate auth response", error=str(exc))
                raise MythicAuthenticationError("Failed to validate auth response") from exc

            self._token = body.access_token
            self._token_expiry = datetime.now() + timedelta(seconds=body.expires_in)
            logger.debug("Got token", expires_in=body.expires_in, expires_at=self._token_expiry)
        return self._token



from functools import cache
from pathlib import Path
from typing import Literal, Union

import rich
from pydantic import ValidationError
from rich.console import Console
from rich.live import Live
from rich.table import Table
from structlog import get_logger

from ..exc import HostedPiValidationError
from ..models.specs import Pi3ServerSpec, Pi4ServerSpec
from ..models.sshkeys import SSHKeySources
from ..pi import Pi
from ..picloud import PiCloud
from . import format


logger = get_logger()
console = Console()


def make_table(*headers: str) -> Table:
    table = Table(show_header=True)
    for header in headers:
        table.add_column(header)
    return table


def validate_server_spec(model: Literal[3, 4], data: dict) -> Union[Pi3ServerSpec, Pi4ServerSpec]:
    if model == 3:
        return Pi3ServerSpec.model_validate(data)
    return Pi4ServerSpec.model_validate(data)


@cache
def get_picloud() -> PiCloud:
    return PiCloud()


def get_pi(name: str) -> Union[Pi, None]:
    cloud = get_picloud()
    return cloud.pis.get(name)


def get_all_pis() -> list[Pi]:
    cloud = get_picloud()
    return list(cloud.pis.values())


def get_pis(names: Union[list[str], None], filter: Union[str, None] = None) -> list[Pi]:
    all_pis = get_all_pis()
    if not names:
        return filter_pis(all_pis, filter)
    all_pi_names = {pi.name for pi in all_pis}
    pis_not_found = [name for name in names if name not in all_pi_names]
    for pi in pis_not_found:
        print_warn(f"Pi server not found: {pi}")
    pis_found = [pi for pi in all_pis if pi.name in names]
    return filter_pis(pis_found, filter)


def filter_pis(pis: list[Pi], filter: Union[str, None]) -> list[Pi]:
    return [pi for pi in pis if filter is None or filter.lower() in pi.name.lower()]


def short_pis_table(pis: list[Pi]):
    table = make_table("Name", "Model", "Memory", "CPU Speed")

    for pi in pis:
        table.add_row(
            pi.name,
            str(pi.model),
            format.memory(pi.memory_gb),
            format.cpu_speed(pi.cpu_speed),
        )
    rich.print(table)


def full_pis_table(pis: list[Pi]):
    headers = [
        "Name",
        "Model",
        "Memory",
        "CPU Speed",
        "NIC Speed",
        "Disk size",
        "Status",
        "Initialised keys",
        "IPv4 SSH port",
        "IPv6 Address",
    ]
    table = Table(*headers)

    with Live(table, console=console, refresh_per_second=4):
        for pi in pis:
            table.add_row(
                pi.name,
                pi.model_full,
                format.memory(pi.memory_gb),
                format.cpu_speed(pi.cpu_speed),
                format.nic_speed(pi.nic_speed),
                format.disk_size(pi.disk_size),
                pi.status,
                format.boolean(pi.initialised_keys),
                str(pi.ipv4_ssh_port),
                pi.ipv6_address.compressed,
            )


def create_pi(
    *,
    model: int,
    disk: int,
    memory_gb: Union[int, None],
    cpu_speed: Union[int, None],
    os_image: Union[str, None],
    wait: bool,
    ssh_key_path: Union[Path, None],
    github_usernames: Union[set[str], None],
    launchpad_usernames: Union[set[str], None],
    full: bool,
    name: Union[str, None] = None,
):
    data = {
        "disk": disk,
        "memory_gb": memory_gb,
        "cpu_speed": cpu_speed,
        "os_image": os_image,
    }
    data = {k: v for k, v in data.items() if v is not None}

    ssh_keys = SSHKeySources(
        ssh_key_path=ssh_key_path,
        github_usernames=github_usernames,
        launchpad_usernames=launchpad_usernames,
    )

    try:
        spec = validate_server_spec(model, data)
    except ValidationError as exc:
        raise HostedPiValidationError(f"Invalid server spec: {exc}") from exc

    cloud = get_picloud()
    pi = cloud.create_pi(name=name, spec=spec, ssh_keys=ssh_keys, wait=wait)

    if full:
        print_success("Server provisioned")
        full_pis_table([pi])
    elif wait:
        print_success("Server provisioned")
        short_pis_table([pi])
    else:
        print_success("Server provision request accepted")


def print_exc(exc: Exception):
    logger.error(f"hostedpi error: {exc}")
    logger.debug("", exc_info=exc)


def print_error(error: str):
    console.print(f"[red]{error}[/red]")


def print_success(message: str):
    console.print(f"[green]{message}[/green]")


def print_warn(message: str):
    console.print(f"[yellow]{message}[/yellow]")
